
#include "gdecoder.h"
#include <iostream>
#include <sstream>
#include <math.h>

using namespace std;
enum operations{none,xflip, yflip, xyexchange, cw,ccw,rot,shift,scale,knive};

float sign(float x)
{
	return x>=0?1:-1;
}


/* a 2d-rotation matrix and a shift. Will be used in future implementations, currently matrizes are set but unused */
void setmatrix(double *m,double *shift,  double m11,double m12,double m21,double m22, double sx,double sy)
{
	m[0]=m11;m[1]=m12;
	m[2]=m21;m[3]=m22;
	shift[0]=sx;
	shift[1]=sy;
}

int main(int argc, char** argv)
{
	
	if(argc==1)
	{
		cerr<<"Usage:\n grecode operation [optional value] gcode.ngc"<<endl;
		cerr<<"Operations:"<<endl;
		cerr<<"xflip, yflip,xyexchange, cw {clockwise 90degree},ccw {counter clockwise}, rot <angle>, scale <factor>, knive <delay>, shift <xshift> <yshift>"<<endl;
		
		return 1;
	}
	operations op=none;
	double rotangle=0,scalefactor=1,knivedelay=0;
	double s[2];
	double m[4];
	for(int i=1;i<argc-1;i++)
	{
		if(string(argv[i])=="xflip")
		{
			op=xflip;
			setmatrix(m,s, -1,0,1,0, 0, 0);
		}
		else
		if(string(argv[i])=="yflip")
		{
			op=yflip;
			setmatrix(m,s, 1,0,-1,0, 0, 0);
		}
		else
		if(string(argv[i])=="xyexchange")
		{
			op=xyexchange;
			setmatrix(m,s, 0,1,1,0, 0, 0);
		}
		else
		if(string(argv[i])=="cw")
		{
			op=cw;
			setmatrix(m,s, 0,1,-1,0, 0, 0);
		}
		else
		if(string(argv[i])=="ccw")
		{
			op=ccw;
			setmatrix(m,s, 0,-1,1,0, 0, 0);
		}
		else
		if(string(argv[i])=="rot")
		{
			op=rot;
			if(i++>=argc)
			{
				cerr<<"rotation angle required"<<endl;
				exit(1);
			}
			stringstream ss(argv[i]);
			ss>>rotangle;
			rotangle*=3.141/180.;
			setmatrix(m,s, cos(rotangle),-sin(rotangle),sin(rotangle),cos(rotangle), 0, 0);
		}
		else
		if(string(argv[i])=="scale")
		{
			op=scale;
			if(i++>=argc)
			{
				cerr<<"scale factor"<<endl;
				exit(1);
			}
			stringstream ss(argv[i]);
			ss>>scalefactor;
			setmatrix(m,s,scalefactor,0,0,scalefactor, 0, 0);
		}
		else
		if(string(argv[i])=="knive")
		{
			op=knive;
			if(i++>=argc)
			{
				cerr<<"knive delay"<<endl;
				exit(1);
			}
			stringstream ss(argv[i]);
			ss>>knivedelay;
			setmatrix(m,s,scalefactor,0,0,scalefactor, 0, 0);
			cerr<<"knivedelay:"<<knivedelay<<endl;
		}
		else
		if(string(argv[i])=="shift")
		{
			op=shift;
			double xs,ys;
			if(i++>=argc)
			{
				cerr<<"x shift required"<<endl;
				exit(1);
			}
			stringstream ss(argv[i]);
			ss>>xs;
			if(i++>=argc)
			{
				cerr<<"y shift required"<<endl;
				exit(1);
			}
			ss.clear();
			ss>>ys;
			setmatrix(m,s, 1,0,0,1, xs, ys);
		}
		else
		{
			cerr<<"I do not understand the option \""<<argv[i]<<"\""<<endl;
		}
	}
	
	
	
	
	GDecoder gd;
	gd.open(argv[argc-1]);
	double lasti;
	//cout.precision( 10 );
	while(gd.good())
	{
		gd.nextLine(); //physical lines in the gcode-file is read, and split into words
		for(int i=0;i<gd.wd.size();i++) //all words of this line
		{
			Word &w=gd.wd[i];
			switch(op)
			{
			case xflip:
				switch(w.type)
				{
				case 'X':
				case 'R':
				case 'I':
					cout<<modify(w,-1);
					break;
				case ' ': //comments and stuff
					cout<<w.text;
					break;
				default:
					cout<<w.type<<w.text<<" ";
				}
				break;
				
			case yflip:
				switch(w.type)
				{
				case 'Y':
				case 'R':
				case 'J':
					cout<<modify(w,-1);
					break;
				case ' ': //comments and stuff
					cout<<w.text;
					break;
				default:
					cout<<w.type<<w.text<<" ";
				}
				break;
				
			case xyexchange:
				switch(w.type)
				{
				case 'Y':
					w.type='X';
					break;
				case 'X':
					w.type='Y';
					break;
				case ' ': //comments and stuff
					cout<<w.text;
					break;
				default:
					;
				}
				if(w.type!=' ')
					cout<<w.type;
				cout<<w.text<<" ";
				break;
			case scale:
				switch(w.type)
				{
				case 'X':
				case 'Y':
					cout<<modify(w,scalefactor);
					break;
				case 'I':
					lasti=w.value;
					//cout.precision( 20 );
					//cout<<modify(w,scalefactor);
					break;
				case 'J':
				{
					float lastj=w.value;
					
					
					double j;
					double dx,dy;
					dx= gd.curPos[0]-gd.lastPos[0];
					dy= gd.curPos[1]-gd.lastPos[1];
					double r=sqrt(SQR(lasti)+SQR(lastj));
					//j=scalefactor*(dy-sign(dy-lastj)*sqrt( SQR(r)-SQR(dx-lasti)  ));
					
					//cout<<"J "<<j<<" ";
					//cout<<";lasti="<<lastj<<" lastj="<<lastj<<"r="<<r<<" dx="<<dx<<" dy="<<dy;
					cout<<"R "<<r*scalefactor<<" ";
					if(1) cout<<"("<<gd.wholeline<<") ";
					break;
				}
				
				case ' ': //comments and stuff
					cout<<w.text;
					break;
				default:
					cout<<w.type<<w.text<<" ";
				}
				break;
			case knive:
				switch(w.type)
				{
				case 'X':
					cout<<w.type<<" "<<"["<<w.text<<"+"<<knivedelay*gd.lastDir[0]<<"]";
					break;
				case 'Y':
					cout<<w.type<<" "<<"["<<w.text<<"+"<<knivedelay*gd.lastDir[1]<<"]";
					break;
				case ' ': //comments and stuff
					cout<<w.text;
					break;
				default:
					cout<<w.type<<w.text<<" ";
				}
				break;
			default:
				cout<<"ERROR; operation not implemented yet:"<<argv[1] <<endl;
				exit(1);
			}
		}
		cout<<endl;
	}
	return 0;
}


